#include <catalogs/icatalog.h>
#include <QScriptEngine>
#include <QDebug>
#include <preferences.h>
#include "artedefinitions.h"

bool addMetadataIfNotEmpty(FilmDetails *film, QVariantMap inputMap, QString fieldName, MetaType internalFieldName, bool isDate) {
    if (!inputMap.value(fieldName).isValid())
        return false;
    QString value = inputMap.value(fieldName).toString();
    if (!isDate) {
        film->m_metadata.insert(internalFieldName, value);
    }
    else {
        // convert "Thu, 09 Apr 2020 05:00:00 +0200" into "25/10/2013 08:30:09"
        if (value.length() > 26 && value.contains(",")) {
            // remove day from "Thu, 09 Apr 2020 05:00:00 +0200"
           value = value.right(value.length() - 5);
        }
        if (value.length() > 21 && value.contains("+")) {
            // remove timezone from "09 Apr 2020 05:00:00 +0200" into "25/10/2013 08:30:09"
            value = value.left(value.length() - 6);
        }

        QString pattern = value.contains("/") ? "dd/MM/yyyy hh:mm:ss" : "dd MMM yyyy hh:mm:ss";

        film->m_metadata.insert(internalFieldName, QLocale(QLocale::English, QLocale::UnitedKingdom)
                                .toDateTime(value, pattern)
                                .toString(Qt::DefaultLocaleShortDate));
    }
    return true;
}

QMap<QString, QVariant> extractJsonMapFromAnswer(QString htmlPage){
    QScriptEngine engine;
    QScriptValue json = engine.evaluate("JSON.parse").call(QScriptValue(),
                                                           QScriptValueList() << QString(htmlPage));
    return json.toVariant().toMap();
}

QList<FilmDetails*> ICatalog::listFilmsFromCatalogAnswer(QString catalogName, const QString& catalogAnswer){
    QString programListProperty = QString("program%1List").arg( Preferences::getInstance()->applicationLanguage().toUpper());
    QList<QVariant> list = extractJsonMapFromAnswer(catalogAnswer).value(programListProperty).toList();

    QList<FilmDetails*> result;
    // http://www.arte.tv/hbbtvv2/services/web/index.php/OPA/v3/streams/069855-019-A/SHOW/fr
    // http://arteptweb.gl-systemhaus.de/am/ptweb/069000/069800/069855-019-A_HQ_0_VF_03336946_MP4-800_AMM-PTWEB_rKCL1CmlPA.mp4
    for (int i = 0; i < list.size(); ++i){
        QMap<QString, QVariant> catalogItem = list.at(i).toMap();
        QString url = catalogItem.value("videoStreamUrl").toString();
        QString title = catalogItem.value("TIT").toString();
        QString arteId = catalogItem.value("PID").toString();

        FilmDetails* newFilm = new FilmDetails(catalogName, title, url, arteId);
        updateArteEpisodeNumber(newFilm);
        addMetadataIfNotEmpty(newFilm, catalogItem, JSON_DESC, Description);
        addMetadataIfNotEmpty(newFilm, catalogItem, JSON_VIEWS, Views);
        addMetadataIfNotEmpty(newFilm, catalogItem, JSON_VIDEO_CHANNEL, Channels);
        result << newFilm;

        QString imageUrl = catalogItem.value("IMG").toMap().value("IUR").toString();
        if (!imageUrl.isEmpty()){
            emit requestImageDownload(newFilm, imageUrl);
        }
    }
    return result;
}

QList<FilmDetails*> ICatalog::sharedListFilmsFromCatalogAnswer(QString responseRootElement, QString catalogName, const QString& catalogAnswer){
    QList<QVariant> list = extractJsonMapFromAnswer(catalogAnswer).value(responseRootElement).toList();
    QList<FilmDetails*> result;
    for (int i = 0; i < list.size(); ++i){
        QMap<QString, QVariant> catalogItem = list.at(i).toMap();
        QString url = QString("https://www.arte.tv/%1/videos/%2")
                .arg(Preferences::getInstance()->applicationLanguage())
                .arg(catalogItem.value("programId").toString());
        QString title = catalogItem.value("title").toString();
        QString arteId = catalogItem.value("programId").toString();

        FilmDetails* newFilm = new FilmDetails(catalogName, title, url, arteId);
        // updateArteEpisodeNumber(newFilm);
        addMetadataIfNotEmpty(newFilm, catalogItem, "shortDescription", Description);

        if (catalogItem.value("durationSeconds").toString().toInt() > 0) {
            // Used by arte date catalog
            newFilm->m_durationInMinutes = catalogItem.value("durationSeconds").toString().toInt() / 60;
        } else {
            // Used by most recent catalog
            newFilm->m_durationInMinutes = catalogItem.value("duration").toInt();
        }

        // V2
        addMetadataIfNotEmpty(newFilm, catalogItem, "videoRightsBegin",   RAW_First_Broadcast, true); // "Mon, 05 Feb 2018 05:00:00 +0100"
        addMetadataIfNotEmpty(newFilm, catalogItem, "videoRightsEnd",      RAW_Available_until, true); // "Mon, 12 Feb 2018 05:00:00 +0100"
        addMetadataIfNotEmpty(newFilm, catalogItem, "director",          Director);
        addMetadataIfNotEmpty(newFilm, catalogItem, "director",          Director);

        extractMainArteMetaFromJsonMap(newFilm, catalogItem);

        result << newFilm;

        QString imageUrl = catalogItem.value("imageUrl").toString();
        if (!imageUrl.isEmpty()){
            emit requestImageDownload(newFilm, imageUrl);
        }


    }
    return result;
}



QString ICatalog::getFilmDetailsUrl(FilmDetails* film){

    QString languageCharacter = Preferences::getInstance()->applicationLanguage().toLower();

    QString jsonUrl = film->m_arteId.isEmpty() ? "" : QString("http://www.arte.tv/hbbtvv2/services/web/index.php/OPA/v3/streams/%0/%1/%2")
                                                 .arg(film->m_arteId)
                                                 .arg(film->m_metadata.value(Kind, "SHOW"))
                                                 .arg(languageCharacter);
    return jsonUrl;
}

void ICatalog::processFilmDetails(FilmDetails* film, QString htmlPage){
    QList<QVariant> streamList = extractJsonMapFromAnswer(htmlPage).value("videoStreams").toList();
    for (int i = 0; i < streamList.size(); ++i){

        QMap<QString, QVariant> streamMap = streamList.at(i).toMap();
        QString streamQuality = streamMap.value("quality").toString();
        // qDebug() << streamQuality;
        if (streamQuality.toLower() == Preferences::getInstance()->selectedQuality()) {
            QString audioType = streamMap.value("audioLabel").toString();
            if (!Preferences::getInstance()->favoriteStreamTypes().contains(audioType)){
                qDebug() << "Add missing stream type in preferences:" << audioType;
                Preferences::getInstance()->addStreamName(audioType);
            }
            // qDebug() << audioType;
            film->m_allStreams[audioType] = streamMap.value("url").toString();
        }
    }
}
